package Shell::EnvImporter::Shell::bad_cons_shell;

use strict;
use warnings;
no warnings 'uninitialized';

use base qw( Shell::EnvImporter::Shell );


#########
sub new {
#########
  return undef;

}

1;
